# estimate.snr
import numpy as np
import audiofile
import matplotlib.pyplot as plt
from scipy.signal.windows import hamming
import argparse


class SNREstimator:
    """Estimate SNR from audio signal using log energy and energy thresholds
    Args:
        input_data (ndarray): Input audio signal
        sample_rate (int): Sampling rate of input audio signal
        window_size (int): Window size in samples
        hop_size (int): Hop size in samples
        
        Returns: 
        object: SNREstimator object
        estimated_snr (float): Estimated SNR in dB, extracted from SNREstimator.estimate_snr()
        
        Usage:
        >>> input_data, sample_rate = audiofile.read('input.wav')
        >>> snr_estimator = SNREstimator(input_data, sample_rate, window_size=320, hop_size=160)
        >>> estimated_snr, log_energies, energy_threshold_low, energy_threshold_high = snr_estimator.estimate_snr() 
    """
    def __init__(self, input_data, sample_rate, window_size=320, hop_size=160):
        self.audio_data = input_data
        self.sample_rate = sample_rate
        self.frame_length = window_size
        self.hop_length = hop_size

    def frame_audio(self, signal):
        num_frames = 1 + (len(signal) - self.frame_length) // self.hop_length
        frames = [signal[i * self.hop_length: (i * self.hop_length) + self.frame_length] for i in range(num_frames)]
        return frames

    def calculate_log_energy(self, frame):
        energy = np.sum(frame ** 2)
        return np.log(energy)

    def calculate_snr(self, energy_high, energy_low):
        return 10 * np.log10(energy_high / energy_low)

    def estimate_snr(self):
        frames = self.frame_audio(self.audio_data)
        log_energies = [self.calculate_log_energy(frame * hamming(self.frame_length)) for frame in frames]
        
        energy_threshold_low = np.percentile(log_energies, 25)  #First quartile
        energy_threshold_high = np.percentile(log_energies, 75)  #Third quartile

        low_energy_frames = [log_energy for log_energy in log_energies if log_energy <= energy_threshold_low]
        high_energy_frames = [log_energy for log_energy in log_energies if log_energy >= energy_threshold_high]
        
        mean_low_energy = np.mean(low_energy_frames)
        mean_high_energy = np.mean(high_energy_frames)
        
        estimated_snr = self.calculate_snr(np.exp(mean_high_energy), np.exp(mean_low_energy))
        return estimated_snr, log_energies, energy_threshold_low, energy_threshold_high

    def plot_energy(self, log_energies, energy_threshold_low, energy_threshold_high):
        plt.figure(figsize=(10, 6))
        plt.plot(log_energies, label='Log Energy')
        plt.axhline(y=energy_threshold_low, color='r', linestyle='--', label='Low Energy Threshold (25th Percentile)')
        plt.axhline(y=energy_threshold_high, color='g', linestyle='--', label='High Energy Threshold (75th Percentile)')
        plt.xlabel('Frame')
        plt.ylabel('Log Energy')
        plt.title('Log Energy and Energy Thresholds')
        plt.legend()
        plt.show()

def main():
    parser = argparse.ArgumentParser(description='Estimate SNR from audio signal', usage='python3 estimate_snr.py -i <input_file> -ws <window_size> -hs <hop_size> -p')
    parser.add_argument('-i', '--input', required=True, help='Input audio file in WAV format')
    parser.add_argument('-ws', '--window_size', type=int, default=int(0.02 * 16000), help='Window size in samples (default: 320)')
    parser.add_argument('-hs', '--hop_size', type=int, default=int(0.01 * 16000), help='Hop size in samples (default: 160)')
    parser.add_argument('-p', '--plot', action='store_true', help='Plot log energy and energy thresholds')
    args = parser.parse_args()

    signal, sr = audiofile.read(args.input)
    snr_estimator = SNREstimator(signal, sr, args.window_size, args.hop_size)
    estimated_snr, log_energies, energy_threshold_low, energy_threshold_high = snr_estimator.estimate_snr()

    print("Estimated SNR:", estimated_snr)
    
    if args.plot:
        snr_estimator.plot_energy(log_energies, energy_threshold_low, energy_threshold_high)

if __name__ == '__main__':
    main()



# featureset.py
import pandas as pd
from nkululeko.util import Util
import nkululeko.glob_conf as glob_conf
import ast


class Featureset:
    name = ""  # designation
    df = None  # pandas dataframe to store the features (and indexed with the data from the sets)
    data_df = None  # dataframe to get audio paths

    def __init__(self, name, data_df):
        self.name = name
        self.data_df = data_df
        self.util = Util("featureset")

    def extract(self):
        pass

    def filter(self):
        # use only the features that are indexed in the target dataframes
        self.df = self.df[self.df.index.isin(self.data_df.index)]
        try:
            # use only some features
            selected_features = ast.literal_eval(glob_conf.config["FEATS"]["features"])
            self.util.debug(f"selecting features: {selected_features}")
            sel_feats_df = pd.DataFrame()
            hit = False
            for feat in selected_features:
                try:
                    sel_feats_df[feat] = self.df[feat]
                    hit = True
                except KeyError:
                    pass
            if hit:
                self.df = sel_feats_df
                self.util.debug(
                    f"new feats shape after selecting features: {self.df.shape}"
                )
        except KeyError:
            pass



""""
A predictor for emotional valence.
Currently based on audEERING's emotional dimension model.
"""
from nkululeko.util import Util
from nkululeko.feature_extractor import FeatureExtractor
import ast
import nkululeko.glob_conf as glob_conf
class ValencePredictor:
    """
    ValencePredictor
    predicting valence with the audEERING emotional dimension model
    
    """
    def __init__(self, df):
        self.df = df
        self.util = Util('valencePredictor')


    def predict(self, split_selection):
        self.util.debug(f'predicting valence for {split_selection} samples')
        feats_name = "_".join(ast.literal_eval(glob_conf.config['DATA']['databases']))
        self.feature_extractor = FeatureExtractor(self.df, ['auddim'], feats_name, split_selection)
        pred_df = self.feature_extractor.extract()
        pred_vals = pred_df.valence * 1000
        return_df = self.df.copy()
        return_df['valence_pred'] = pred_vals.astype('int')/1000

        return return_df




# resample.py
# change the sampling rate for train and test splits

from nkululeko.experiment import Experiment
import configparser
from nkululeko.util import Util
from  nkululeko.constants import VERSION
import argparse
import os
import pandas as pd
from nkululeko.augmenting.resampler import Resampler

def main(src_dir):
    parser = argparse.ArgumentParser(description='Call the nkululeko framework.')
    parser.add_argument('--config', default='exp.ini', help='The base configuration')
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config    
    else:
        config_file = f'{src_dir}/exp.ini'

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f'ERROR: no such file: {config_file}')
        exit()
        
    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    # create a new experiment
    expr = Experiment(config)
    util = Util('resample')
    util.debug(f'running {expr.name} from config {config_file}, nkululeko version {VERSION}')

    if util.config_val('EXP', 'no_warnings', False):
        import warnings
        warnings.filterwarnings("ignore")

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(f'train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}')

    sample_selection = util.config_val('RESAMPLE', 'sample_selection', 'all')
    if sample_selection=='all':
        df = pd.concat([expr.df_train, expr.df_test])
    elif sample_selection=='train':
        df = expr.df_train
    elif sample_selection=='test':
        df = expr.df_test
    else:
        util.error(f'unknown selection specifier {sample_selection}, should be [all | train | test]')
    util.debug(f'resampling {sample_selection}: {df.shape[0]} samples') 
    rs = Resampler(df)
    rs.resample()
    print('DONE')


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd) # use this if you want to state the config file path on command line


# segment.py
# segment data splits

from nkululeko.experiment import Experiment
import configparser
from nkululeko.util import Util
from  nkululeko.constants import VERSION
import argparse
import os
import pandas as pd 

def main(src_dir):
    parser = argparse.ArgumentParser(description='Call the nkululeko framework.')
    parser.add_argument('--config', default='exp.ini', help='The base configuration')
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config    
    else:
        config_file = f'{src_dir}/exp.ini'

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f'ERROR: no such file: {config_file}')
        exit()
        
    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    # create a new experiment
    expr = Experiment(config)
    util = Util('segment')
    util.debug(f'running {expr.name}, nkululeko version {VERSION}')

    if util.config_val('EXP', 'no_warnings', False):
        import warnings
        warnings.filterwarnings("ignore")

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(f'train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}')

    # segment
    segment_target = util.config_val('SEGMENT', 'target', '_seg')

    segmenter = util.config_val('SEGMENT', 'method', 'silero')
    sample_selection = util.config_val('SEGMENT', 'sample_selection', 'all')
    if sample_selection=='all':
        df = pd.concat([expr.df_train, expr.df_test])
    elif sample_selection=='train':
        df = expr.df_train
    elif sample_selection=='test':
        df = expr.df_test
    else:
        util.error(f'unknown segmentation selection specifier {sample_selection}, should be [all | train | test]')
    util.debug(f'segmenting {sample_selection}: {df.shape[0]} samples with {segmenter}') 
    if segmenter=='silero':
        from nkululeko.segmenting.seg_silero import Silero_segmenter
        segmenter = Silero_segmenter()
        df_seg = segmenter.segment_dataframe(df)

    else:
        util.error(f'unkown segmenter: {segmenter}')

    def calc_dur(x):
        from datetime import datetime
        starts = x[1]
        ends = x[2]
        return (ends - starts).total_seconds()
    df_seg['duration'] = df_seg.index.to_series().map(lambda x:calc_dur(x)) 
    if "duration" not in df.columns:
        df['duration'] = df.index.to_series().map(lambda x:calc_dur(x)) 
    num_before = df.shape[0]
    num_after = df_seg.shape[0]
    # plot distributions
    from nkululeko.plots import Plots
    plots = Plots()
    plots.plot_durations(df, 'original_durations', sample_selection)
    plots.plot_durations(df_seg, 'segmented_durations', sample_selection)
    print('')
    # remove encoded labels
    target = util.config_val('DATA', 'target', 'emotion')
    if 'class_label' in df_seg.columns:
        df_seg = df_seg.drop(columns=[target])
        df_seg = df_seg.rename(columns={'class_label':target})
    # save file
    dataname = '_'.join(expr.datasets.keys())
    name = f'{dataname}{segment_target}'
    df_seg.to_csv(f'{expr.data_dir}/{name}.csv')
    util.debug(f'saved {name}.csv to {expr.data_dir}, {num_after} samples (was {num_before})')
    print('DONE')

def get_segmentation(file):
#    print(f'segmenting {file[0]}')
    print('.', end='')
    wav = read_audio(file[0], sampling_rate=SAMPLING_RATE)
    speech_timestamps = get_speech_timestamps(wav, vad_model, sampling_rate=SAMPLING_RATE)
    files, starts, ends = [], [], []
    for entry in speech_timestamps:
        start = float(entry['start']/1000.0)
        end = float(entry['end']/1000.0)
        files.append(file[0])
        starts.append(start)
        ends.append(end)
    seg_index = segmented_index(files, starts, ends)
    return seg_index

def segment_dataframe(df):
    dfs = []
    for file, values in df.iterrows():
        index = get_segmentation(file)
        dfs.append(
            pd.DataFrame(
                values.to_dict(),
                index,
            )
        )
    return audformat.utils.concat(dfs)


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd) # use this if you want to state the config file path on command line


#!/usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#                                                                         #
#  Praat Script Syllable Nuclei                                           #
#  Copyright (C) 2008  Nivja de Jong and Ton Wempe                        #
#                                                                         #
#    This program is free software: you can redistribute it and/or modify #
#    it under the terms of the GNU General Public License as published by #
#    the Free Software Foundation, either version 3 of the License, or    #
#    (at your option) any later version.                                  #
#                                                                         #
#    This program is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        #
#    GNU General Public License for more details.                         #
#                                                                         #
#    You should have received a copy of the GNU General Public License    #
#    along with this program.  If not, see http://www.gnu.org/licenses/   #
#                                                                         #
###########################################################################
#
# modified 2010.09.17 by Hugo Quené, Ingrid Persoon, & Nivja de Jong
# Overview of changes:
# + change threshold-calculator: rather than using median, use the almost maximum
#     minus 25dB. (25 dB is in line with the standard setting to detect silence
#     in the "To TextGrid (silences)" function.
#     Almost maximum (.99 quantile) is used rather than maximum to avoid using
#     irrelevant non-speech sound-bursts.
# + add silence-information to calculate articulation rate and ASD (average syllable
#     duration.
#     NB: speech rate = number of syllables / total time
#         articulation rate = number of syllables / phonation time
# + remove max number of syllable nuclei
# + refer to objects by unique identifier, not by name
# + keep track of all created intermediate objects, select these explicitly,
#     then Remove
# + provide summary output in Info window
# + do not save TextGrid-file but leave it in Object-window for inspection
#     (if requested in startup-form)
# + allow Sound to have starting time different from zero
#      for Sound objects created with Extract (preserve times)
# + programming of checking loop for mindip adjusted
#      in the orig version, precedingtime was not modified if the peak was rejected !!
#      var precedingtime and precedingint renamed to currenttime and currentint
#
# + bug fixed concerning summing total pause, feb 28th 2011
###########################################################################


# counts syllables of all sound utterances in a directory
# NB unstressed syllables are sometimes overlooked
# NB filter sounds that are quite noisy beforehand
# NB use Silence threshold (dB) = -25 (or -20?)
# NB use Minimum dip between peaks (dB) = between 2-4 (you can first try;
#                                                      For clean and filtered: 4)
#
#
# Translated to Python in 2019 by David Feinberg
# I changed all the variable names so they are human readable



import math
import pandas as pd
import parselmouth
from tqdm import tqdm

from glob import glob
from parselmouth.praat import call


def speech_rate(filename):
    silencedb = -25
    mindip = 2
    minpause = 0.3
    sound = parselmouth.Sound(filename)
    originaldur = sound.get_total_duration()
    intensity = sound.to_intensity(50)
    start = call(intensity, "Get time from frame number", 1)
    nframes = call(intensity, "Get number of frames")
    end = call(intensity, "Get time from frame number", nframes)
    min_intensity = call(intensity, "Get minimum", 0, 0, "Parabolic")
    max_intensity = call(intensity, "Get maximum", 0, 0, "Parabolic")

    # get .99 quantile to get maximum (without influence of non-speech sound bursts)
    max_99_intensity = call(intensity, "Get quantile", 0, 0, 0.99)

    # estimate Intensity threshold
    threshold = max_99_intensity + silencedb
    threshold2 = max_intensity - max_99_intensity
    threshold3 = silencedb - threshold2
    if threshold < min_intensity:
        threshold = min_intensity

    # get pauses (silences) and speakingtime
    textgrid = call(intensity, "To TextGrid (silences)", threshold3, minpause, 0.1, "silent", "sounding")
    silencetier = call(textgrid, "Extract tier", 1)
    silencetable = call(silencetier, "Down to TableOfReal", "sounding")
    npauses = call(silencetable, "Get number of rows")
    speakingtot = 0
    for ipause in range(npauses):
        pause = ipause + 1
        beginsound = call(silencetable, "Get value", pause, 1)
        endsound = call(silencetable, "Get value", pause, 2)
        speakingdur = endsound - beginsound
        speakingtot += speakingdur

    intensity_matrix = call(intensity, "Down to Matrix")
    # sndintid = sound_from_intensity_matrix
    sound_from_intensity_matrix = call(intensity_matrix, "To Sound (slice)", 1)
    # use total duration, not end time, to find out duration of intdur (intensity_duration)
    # in order to allow nonzero starting times.
    intensity_duration = call(sound_from_intensity_matrix, "Get total duration")
    intensity_max = call(sound_from_intensity_matrix, "Get maximum", 0, 0, "Parabolic")
    point_process = call(sound_from_intensity_matrix, "To PointProcess (extrema)", "Left", "yes", "no", "Sinc70")
    # estimate peak positions (all peaks)
    numpeaks = call(point_process, "Get number of points")
    t = [call(point_process, "Get time from index", i + 1) for i in range(numpeaks)]

    # fill array with intensity values
    timepeaks = []
    peakcount = 0
    intensities = []
    for i in range(numpeaks):
        value = call(sound_from_intensity_matrix, "Get value at time", t[i], "Cubic")
        if value > threshold:
            peakcount += 1
            intensities.append(value)
            timepeaks.append(t[i])

    # fill array with valid peaks: only intensity values if preceding
    # dip in intensity is greater than mindip
    validpeakcount = 0
    currenttime = timepeaks[0]
    currentint = intensities[0]
    validtime = []

    for p in range(peakcount - 1):
        following = p + 1
        followingtime = timepeaks[p + 1]
        dip = call(intensity, "Get minimum", currenttime, timepeaks[p + 1], "None")
        diffint = abs(currentint - dip)
        if diffint > mindip:
            validpeakcount += 1
            validtime.append(timepeaks[p])
        currenttime = timepeaks[following]
        currentint = call(intensity, "Get value at time", timepeaks[following], "Cubic")

    # Look for only voiced parts
    pitch = sound.to_pitch_ac(0.02, 30, 4, False, 0.03, 0.25, 0.01, 0.35, 0.25, 450)
    voicedcount = 0
    voicedpeak = []

    for time in range(validpeakcount):
        querytime = validtime[time]
        whichinterval = call(textgrid, "Get interval at time", 1, querytime)
        whichlabel = call(textgrid, "Get label of interval", 1, whichinterval)
        value = pitch.get_value_at_time(querytime) 
        if not math.isnan(value):
            if whichlabel == "sounding":
                voicedcount += 1
                voicedpeak.append(validtime[time])

    # calculate time correction due to shift in time for Sound object versus
    # intensity object
    timecorrection = originaldur / intensity_duration

    # Insert voiced peaks in TextGrid
    call(textgrid, "Insert point tier", 1, "syllables")
    for i in range(len(voicedpeak)):
        position = (voicedpeak[i] * timecorrection)
        call(textgrid, "Insert point", 1, position, "")

    # return results
    speakingrate = voicedcount / originaldur
    articulationrate = voicedcount / speakingtot
    npause = npauses - 1
    try:
        asd = speakingtot / voicedcount
    except ZeroDivisionError:
        asd = 0
        print('caught zero division')
    speechrate_dictionary = {'file':file,
                             'nsyll':voicedcount,
                             'npause': npause,
                             'dur(s)':originaldur,
                             'phonationtime(s)':intensity_duration,
                             'speechrate(nsyll / dur)': speakingrate,
                             "articulation rate(nsyll / phonationtime)":articulationrate,
                             "ASD(speakingtime / nsyll)":asd}
    return speechrate_dictionary


def get_files():
    audio_dir = '/home/audeering.local/fburkhardt/audb/projectsmile-speakingrate-agent/3.0.0/fe182b91/audio/'
    files = glob(f'{audio_dir}/*.wav')
    files.extend(glob('test_voices/*.mp3'))
    files.extend(glob('test_voices/*.ogg'))
    files.extend(glob('test_voices/*.aiff'))
    files.extend(glob('test_voices/*.aifc'))
    files.extend(glob('test_voices/*.au'))
    files.extend(glob('test_voices/*.nist'))
    files.extend(glob('test_voices/*.flac'))
    return files


if __name__ == "__main__":
    files = get_files()
    import audformat
    df_test = audformat.utils.read_csv('speechrate_tests.csv')
    files = df_test.index.values
    cols = ['file', 'nsyll', 'npause', 'dur(s)', 'phonationtime(s)', 'speechrate(nsyll / dur)', 'articulation '
            'rate(nsyll / phonationtime)', 'ASD(speakingtime / nsyll)']
    datalist = []
    for index, file in enumerate(tqdm(files)):
        # print(f'processing {file}')
        speechrate_dictionary = speech_rate(file)
        datalist.append(speechrate_dictionary)
    df = pd.DataFrame(datalist)
    print(f'processed {index} files')
    df.to_csv('speechrate_data.csv', index=0)


"""
This is a copy of David R. Feinberg's Praat scripts
https://github.com/drfeinberg/PraatScripts
taken June 23rd 2022
"""

#!/usr/bin/env python3
import numpy as np
import pandas as pd
import math
from tqdm import tqdm
import parselmouth
import statistics
from nkululeko.util import Util
import audiofile
from parselmouth.praat import call
from scipy.stats.mstats import zscore
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler


# This is the function to measure source acoustics using default male parameters.


def measurePitch(voiceID, f0min, f0max, unit):
    sound = parselmouth.Sound(voiceID)  # read the sound
    duration = call(sound, "Get total duration")  # duration
    pitch = call(sound, "To Pitch", 0.0, f0min, f0max)  # create a praat pitch object
    meanF0 = call(pitch, "Get mean", 0, 0, unit)  # get mean pitch
    stdevF0 = call(
        pitch, "Get standard deviation", 0, 0, unit
    )  # get standard deviation
    harmonicity = call(sound, "To Harmonicity (cc)", 0.01, f0min, 0.1, 1.0)
    hnr = call(harmonicity, "Get mean", 0, 0)
    pointProcess = call(sound, "To PointProcess (periodic, cc)", f0min, f0max)
    localJitter = call(pointProcess, "Get jitter (local)", 0, 0, 0.0001, 0.02, 1.3)
    localabsoluteJitter = call(
        pointProcess, "Get jitter (local, absolute)", 0, 0, 0.0001, 0.02, 1.3
    )
    rapJitter = call(pointProcess, "Get jitter (rap)", 0, 0, 0.0001, 0.02, 1.3)
    ppq5Jitter = call(pointProcess, "Get jitter (ppq5)", 0, 0, 0.0001, 0.02, 1.3)
    ddpJitter = call(pointProcess, "Get jitter (ddp)", 0, 0, 0.0001, 0.02, 1.3)
    localShimmer = call(
        [sound, pointProcess], "Get shimmer (local)", 0, 0, 0.0001, 0.02, 1.3, 1.6
    )
    localdbShimmer = call(
        [sound, pointProcess], "Get shimmer (local_dB)", 0, 0, 0.0001, 0.02, 1.3, 1.6
    )
    apq3Shimmer = call(
        [sound, pointProcess], "Get shimmer (apq3)", 0, 0, 0.0001, 0.02, 1.3, 1.6
    )
    aqpq5Shimmer = call(
        [sound, pointProcess], "Get shimmer (apq5)", 0, 0, 0.0001, 0.02, 1.3, 1.6
    )
    apq11Shimmer = call(
        [sound, pointProcess], "Get shimmer (apq11)", 0, 0, 0.0001, 0.02, 1.3, 1.6
    )
    ddaShimmer = call(
        [sound, pointProcess], "Get shimmer (dda)", 0, 0, 0.0001, 0.02, 1.3, 1.6
    )

    return (
        duration,
        meanF0,
        stdevF0,
        hnr,
        localJitter,
        localabsoluteJitter,
        rapJitter,
        ppq5Jitter,
        ddpJitter,
        localShimmer,
        localdbShimmer,
        apq3Shimmer,
        aqpq5Shimmer,
        apq11Shimmer,
        ddaShimmer,
    )


# ## This function measures formants at each glottal pulse
#
# Puts, D. A., Apicella, C. L., & Cárdenas, R. A. (2012). Masculine voices signal men's threat potential in forager and industrial societies. Proceedings of the Royal Society of London B: Biological Sciences, 279(1728), 601-609.
#
# Adapted from: DOI 10.17605/OSF.IO/K2BHS
# This function measures formants using Formant Position formula
# def measureFormants(sound, wave_file, f0min,f0max):
def measureFormants(sound, f0min, f0max):
    sound = parselmouth.Sound(sound)  # read the sound
    #    pitch = call(sound, "To Pitch (cc)", 0, f0min, 15, 'no', 0.03, 0.45, 0.01, 0.35, 0.14, f0max)
    pointProcess = call(sound, "To PointProcess (periodic, cc)", f0min, f0max)

    formants = call(sound, "To Formant (burg)", 0.0025, 5, 5000, 0.025, 50)
    numPoints = call(pointProcess, "Get number of points")

    f1_list = []
    f2_list = []
    f3_list = []
    f4_list = []

    # Measure formants only at glottal pulses
    for point in range(0, numPoints):
        point += 1
        t = call(pointProcess, "Get time from index", point)
        f1 = call(formants, "Get value at time", 1, t, "Hertz", "Linear")
        f2 = call(formants, "Get value at time", 2, t, "Hertz", "Linear")
        f3 = call(formants, "Get value at time", 3, t, "Hertz", "Linear")
        f4 = call(formants, "Get value at time", 4, t, "Hertz", "Linear")
        f1_list.append(f1)
        f2_list.append(f2)
        f3_list.append(f3)
        f4_list.append(f4)

    f1_list = [f1 for f1 in f1_list if str(f1) != "nan"]
    f2_list = [f2 for f2 in f2_list if str(f2) != "nan"]
    f3_list = [f3 for f3 in f3_list if str(f3) != "nan"]
    f4_list = [f4 for f4 in f4_list if str(f4) != "nan"]

    # calculate mean formants across pulses
    f1_mean = statistics.mean(f1_list)
    f2_mean = statistics.mean(f2_list)
    f3_mean = statistics.mean(f3_list)
    f4_mean = statistics.mean(f4_list)

    # calculate median formants across pulses, this is what is used in all subsequent calcualtions
    # you can use mean if you want, just edit the code in the boxes below to replace median with mean
    f1_median = statistics.median(f1_list)
    f2_median = statistics.median(f2_list)
    f3_median = statistics.median(f3_list)
    f4_median = statistics.median(f4_list)

    return (
        f1_mean,
        f2_mean,
        f3_mean,
        f4_mean,
        f1_median,
        f2_median,
        f3_median,
        f4_median,
    )


# ## This function runs a 2-factor Principle Components Analysis (PCA) on Jitter and Shimmer


def runPCA(df):
    # z-score the Jitter and Shimmer measurements
    measures = [
        "localJitter",
        "localabsoluteJitter",
        "rapJitter",
        "ppq5Jitter",
        "ddpJitter",
        "localShimmer",
        "localdbShimmer",
        "apq3Shimmer",
        "apq5Shimmer",
        "apq11Shimmer",
        "ddaShimmer",
    ]
    x = df.loc[:, measures].values
    # f = open('x.pickle', 'wb')
    # pickle.dump(x, f)
    # f.close()

    x = StandardScaler().fit_transform(x)
    if np.any(np.isnan(x)):
        print(f"Warning: {np.count_nonzero(np.isnan(x))} Nans in x, replacing with 0")
        x[np.isnan(x)] = 0
    if np.any(np.isfinite(x)):
        print(f"Warning: {np.count_nonzero(np.isfinite(x))} infinite in x")

    # PCA
    pca = PCA(n_components=2)
    principalComponents = pca.fit_transform(x)
    if np.any(np.isnan(principalComponents)):
        print("pc is nan")
        print(f"count: {np.count_nonzero(np.isnan(principalComponents))}")
        print(principalComponents)
        principalComponents = np.nan_to_num(principalComponents)

    principalDf = pd.DataFrame(
        data=principalComponents, columns=["JitterPCA", "ShimmerPCA"]
    )

    return principalDf


# ## This block of code runs the above functions on all of the '.wav' files in the /audio folder


def compute_features(file_index):
    # create lists to put the results
    file_list = []
    duration_list = []
    mean_F0_list = []
    sd_F0_list = []
    hnr_list = []
    localJitter_list = []
    localabsoluteJitter_list = []
    rapJitter_list = []
    ppq5Jitter_list = []
    ddpJitter_list = []
    localShimmer_list = []
    localdbShimmer_list = []
    apq3Shimmer_list = []
    aqpq5Shimmer_list = []
    apq11Shimmer_list = []
    ddaShimmer_list = []
    f1_mean_list = []
    f2_mean_list = []
    f3_mean_list = []
    f4_mean_list = []
    f1_median_list = []
    f2_median_list = []
    f3_median_list = []
    f4_median_list = []
    # Go through all the wave files in the folder and measure all the acoustics
    #    for i, wave_file in enumerate(file_list):
    for idx, (wave_file, start, end) in enumerate(tqdm(file_index.to_list())):
        signal, sampling_rate = audiofile.read(
            wave_file,
            offset=start.total_seconds(),
            duration=(end - start).total_seconds(),
            always_2d=True,
        )
        sound = parselmouth.Sound(values=signal, sampling_frequency=sampling_rate)
        (
            duration,
            meanF0,
            stdevF0,
            hnr,
            localJitter,
            localabsoluteJitter,
            rapJitter,
            ppq5Jitter,
            ddpJitter,
            localShimmer,
            localdbShimmer,
            apq3Shimmer,
            aqpq5Shimmer,
            apq11Shimmer,
            ddaShimmer,
        ) = measurePitch(sound, 75, 300, "Hertz")
        (
            f1_mean,
            f2_mean,
            f3_mean,
            f4_mean,
            f1_median,
            f2_median,
            f3_median,
            f4_median,
        ) = measureFormants(sound, 75, 300)
        #        file_list.append(wave_file) # make an ID list
        duration_list.append(duration)  # make duration list
        mean_F0_list.append(meanF0)  # make a mean F0 list
        sd_F0_list.append(stdevF0)  # make a sd F0 list
        hnr_list.append(hnr)  # add HNR data

        # add raw jitter and shimmer measures
        localJitter_list.append(localJitter)
        localabsoluteJitter_list.append(localabsoluteJitter)
        rapJitter_list.append(rapJitter)
        ppq5Jitter_list.append(ppq5Jitter)
        ddpJitter_list.append(ddpJitter)
        localShimmer_list.append(localShimmer)
        localdbShimmer_list.append(localdbShimmer)
        apq3Shimmer_list.append(apq3Shimmer)
        aqpq5Shimmer_list.append(aqpq5Shimmer)
        apq11Shimmer_list.append(apq11Shimmer)
        ddaShimmer_list.append(ddaShimmer)

        # add the formant data
        f1_mean_list.append(f1_mean)
        f2_mean_list.append(f2_mean)
        f3_mean_list.append(f3_mean)
        f4_mean_list.append(f4_mean)
        f1_median_list.append(f1_median)
        f2_median_list.append(f2_median)
        f3_median_list.append(f3_median)
        f4_median_list.append(f4_median)

    # ## This block of code adds all of that data we just generated to a Pandas data frame
    # Add the data to Pandas
    df = pd.DataFrame(
        np.column_stack(
            [
                duration_list,
                mean_F0_list,
                sd_F0_list,
                hnr_list,
                localJitter_list,
                localabsoluteJitter_list,
                rapJitter_list,
                ppq5Jitter_list,
                ddpJitter_list,
                localShimmer_list,
                localdbShimmer_list,
                apq3Shimmer_list,
                aqpq5Shimmer_list,
                apq11Shimmer_list,
                ddaShimmer_list,
                f1_mean_list,
                f2_mean_list,
                f3_mean_list,
                f4_mean_list,
                f1_median_list,
                f2_median_list,
                f3_median_list,
                f4_median_list,
            ]
        ),
        columns=[
            "duration",
            "meanF0Hz",
            "stdevF0Hz",
            "HNR",
            "localJitter",
            "localabsoluteJitter",
            "rapJitter",
            "ppq5Jitter",
            "ddpJitter",
            "localShimmer",
            "localdbShimmer",
            "apq3Shimmer",
            "apq5Shimmer",
            "apq11Shimmer",
            "ddaShimmer",
            "f1_mean",
            "f2_mean",
            "f3_mean",
            "f4_mean",
            "f1_median",
            "f2_median",
            "f3_median",
            "f4_median",
        ],
    )

    # add pca data
    pcaData = runPCA(df)  # Run jitter and shimmer PCA
    df = pd.concat([df, pcaData], axis=1)  # Add PCA data
    # reload the data so it's all numbers
    df.to_csv("processed_results.csv", index=False)
    df = pd.read_csv("processed_results.csv", header=0)
    #    df.sort_values('voiceID').head(20)
    # ## Next we calculate the vocal-tract length estimates

    # ### Formant position
    #  Puts, D. A., Apicella, C. L., & Cárdenas, R. A. (2012). Masculine voices signal men's threat potential in forager and industrial societies. Proceedings of the Royal Society of London B: Biological Sciences, 279(1728), 601-609.

    df["pF"] = (
        zscore(df.f1_median)
        + zscore(df.f2_median)
        + zscore(df.f3_median)
        + zscore(df.f4_median)
    ) / 4

    # ### Formant Dispersion
    # Fitch, W. T. (1997). Vocal tract length and formant frequency dispersion correlate with body size in rhesus macaques. The Journal of the Acoustical Society of America, 102(2), 1213-1222.

    df["fdisp"] = (df["f4_median"] - df["f1_median"]) / 3

    # ### Fn (Average Formant)
    # Pisanski, K., & Rendall, D. (2011). The prioritization of voice fundamental frequency or formants in listeners’ assessments of speaker size, masculinity, and attractiveness. The Journal of the Acoustical Society of America, 129(4), 2201-2212.

    df["avgFormant"] = (
        df["f1_median"] + df["f2_median"] + df["f3_median"] + df["f4_median"]
    ) / 4

    # ### MFF
    # Smith, D. R., & Patterson, R. D. (2005). The interaction of glottal-pulse rate and vocal-tract length in judgements of speaker size, sex, and age. The Journal of the Acoustical Society of America, 118(5), 3177-3186.

    df["mff"] = (
        df["f1_median"] * df["f2_median"] * df["f3_median"] * df["f4_median"]
    ) ** 0.25

    # ### Fitch VTL
    # Fitch, W. T. (1997). Vocal tract length and formant frequency dispersion correlate with body size in rhesus macaques. The Journal of the Acoustical Society of America, 102(2), 1213-1222.

    # reload the data again
    # df.to_csv("processed_results.csv", index=False)
    # df = pd.read_csv('processed_results.csv', header=0)

    df["fitch_vtl"] = (
        (1 * (35000 / (4 * df["f1_median"])))
        + (3 * (35000 / (4 * df["f2_median"])))
        + (5 * (35000 / (4 * df["f3_median"])))
        + (7 * (35000 / (4 * df["f4_median"])))
    ) / 4

    # ### $\Delta$F
    # Reby,D.,& McComb,K.(2003). Anatomical constraints generate honesty: acoustic cues to age and weight in the roars of red deer stags. Animal Behaviour, 65, 519e-530.

    xysum = (
        (0.5 * df["f1_median"])
        + (1.5 * df["f2_median"])
        + (2.5 * df["f3_median"])
        + (3.5 * df["f4_median"])
    )
    xsquaredsum = (0.5**2) + (1.5**2) + (2.5**2) + (3.5**2)
    df["delta_f"] = xysum / xsquaredsum

    # ### VTL($\Delta$F)
    # Reby,D.,&McComb,K.(2003).Anatomical constraints generate honesty: acoustic cues to age and weight in the roars of red deer stags. Animal Behaviour, 65, 519e-530.

    df["vtl_delta_f"] = 35000 / (2 * df["delta_f"])

    print("Now extracting speech rate parameters...")

    df_speechrate = get_speech_rate(file_index)
    print("")

    return df.join(df_speechrate)


"""
Speech rate script taken from https://github.com/drfeinberg/PraatScripts
on 25/05/23
"""


def get_speech_rate(file_index):
    cols = [
        "nsyll",
        "npause",
        "dur(s)",
        "phonationtime(s)",
        "speechrate(nsyll / dur)",
        "articulation " "rate(nsyll / phonationtime)",
        "ASD(speakingtime / nsyll)",
    ]
    datalist = []
    for idx, (wave_file, start, end) in enumerate(tqdm(file_index.to_list())):
        signal, sampling_rate = audiofile.read(
            wave_file,
            offset=start.total_seconds(),
            duration=(end - start).total_seconds(),
            always_2d=True,
        )
        sound = parselmouth.Sound(values=signal, sampling_frequency=sampling_rate)
        # print(f'processing {file}')
        speechrate_dictionary = speech_rate(sound)
        datalist.append(speechrate_dictionary)
    df = pd.DataFrame(datalist)
    return df


def speech_rate(sound):
    silencedb = -25
    mindip = 2
    minpause = 0.3
    originaldur = sound.get_total_duration()
    intensity = sound.to_intensity(50)
    start = call(intensity, "Get time from frame number", 1)
    nframes = call(intensity, "Get number of frames")
    end = call(intensity, "Get time from frame number", nframes)
    min_intensity = call(intensity, "Get minimum", 0, 0, "Parabolic")
    max_intensity = call(intensity, "Get maximum", 0, 0, "Parabolic")

    # get .99 quantile to get maximum (without influence of non-speech sound bursts)
    max_99_intensity = call(intensity, "Get quantile", 0, 0, 0.99)

    # estimate Intensity threshold
    threshold = max_99_intensity + silencedb
    threshold2 = max_intensity - max_99_intensity
    threshold3 = silencedb - threshold2
    if threshold < min_intensity:
        threshold = min_intensity

    # get pauses (silences) and speakingtime
    textgrid = call(
        intensity,
        "To TextGrid (silences)",
        threshold3,
        minpause,
        0.1,
        "silent",
        "sounding",
    )
    silencetier = call(textgrid, "Extract tier", 1)
    silencetable = call(silencetier, "Down to TableOfReal", "sounding")
    npauses = call(silencetable, "Get number of rows")
    speakingtot = 0
    for ipause in range(npauses):
        pause = ipause + 1
        beginsound = call(silencetable, "Get value", pause, 1)
        endsound = call(silencetable, "Get value", pause, 2)
        speakingdur = endsound - beginsound
        speakingtot += speakingdur

    intensity_matrix = call(intensity, "Down to Matrix")
    # sndintid = sound_from_intensity_matrix
    sound_from_intensity_matrix = call(intensity_matrix, "To Sound (slice)", 1)
    # use total duration, not end time, to find out duration of intdur (intensity_duration)
    # in order to allow nonzero starting times.
    intensity_duration = call(sound_from_intensity_matrix, "Get total duration")
    intensity_max = call(sound_from_intensity_matrix, "Get maximum", 0, 0, "Parabolic")
    point_process = call(
        sound_from_intensity_matrix,
        "To PointProcess (extrema)",
        "Left",
        "yes",
        "no",
        "Sinc70",
    )
    # estimate peak positions (all peaks)
    numpeaks = call(point_process, "Get number of points")
    t = [call(point_process, "Get time from index", i + 1) for i in range(numpeaks)]

    # fill array with intensity values
    timepeaks = []
    peakcount = 0
    intensities = []
    for i in range(numpeaks):
        value = call(sound_from_intensity_matrix, "Get value at time", t[i], "Cubic")
        if value > threshold:
            peakcount += 1
            intensities.append(value)
            timepeaks.append(t[i])

    # fill array with valid peaks: only intensity values if preceding
    # dip in intensity is greater than mindip
    validpeakcount = 0
    currenttime = timepeaks[0]
    currentint = intensities[0]
    validtime = []

    for p in range(peakcount - 1):
        following = p + 1
        followingtime = timepeaks[p + 1]
        dip = call(intensity, "Get minimum", currenttime, timepeaks[p + 1], "None")
        diffint = abs(currentint - dip)
        if diffint > mindip:
            validpeakcount += 1
            validtime.append(timepeaks[p])
        currenttime = timepeaks[following]
        currentint = call(intensity, "Get value at time", timepeaks[following], "Cubic")

    # Look for only voiced parts
    pitch = sound.to_pitch_ac(0.02, 30, 4, False, 0.03, 0.25, 0.01, 0.35, 0.25, 450)
    voicedcount = 0
    voicedpeak = []

    for time in range(validpeakcount):
        querytime = validtime[time]
        whichinterval = call(textgrid, "Get interval at time", 1, querytime)
        whichlabel = call(textgrid, "Get label of interval", 1, whichinterval)
        value = pitch.get_value_at_time(querytime)
        if not math.isnan(value):
            if whichlabel == "sounding":
                voicedcount += 1
                voicedpeak.append(validtime[time])

    # calculate time correction due to shift in time for Sound object versus
    # intensity object
    timecorrection = originaldur / intensity_duration

    # Insert voiced peaks in TextGrid
    call(textgrid, "Insert point tier", 1, "syllables")
    for i in range(len(voicedpeak)):
        position = voicedpeak[i] * timecorrection
        call(textgrid, "Insert point", 1, position, "")

    # return results
    speakingrate = voicedcount / originaldur
    articulationrate = voicedcount / speakingtot
    npause = npauses - 1
    try:
        asd = speakingtot / voicedcount
    except ZeroDivisionError:
        asd = 0
        print("caught zero division")
    speechrate_dictionary = {
        "nsyll": voicedcount,
        "npause": npause,
        "dur(s)": originaldur,
        "phonationtime(s)": intensity_duration,
        "speechrate(nsyll / dur)": speakingrate,
        "articulation rate(nsyll / phonationtime)": articulationrate,
        "ASD(speakingtime / nsyll)": asd,
    }
    return speechrate_dictionary



import torch
from torch.nn.functional import one_hot, softmax


class SoftF1Loss(torch.nn.Module):

    """differentiable F1 loss, adapted from
    https://gist.github.com/SuperShinyEyes/dcc68a08ff8b615442e3bc6a9b55a354

    written by Uwe Reichel"""

    def __init__(self, weight=None, epsilon=1e-7, num_classes=-1):
        super().__init__()
        self.weight = weight
        self.epsilon = epsilon
        self.num_classes = num_classes

    def forward(self, y_pred, y_true):
        assert y_pred.ndim == 2
        assert y_true.ndim == 1

        y_true = one_hot(y_true, self.num_classes).to(torch.float32)
        y_pred = softmax(y_pred, dim=1)

        tp = (y_true * y_pred).sum(dim=0).to(torch.float32)
        fp = ((1 - y_true) * y_pred).sum(dim=0).to(torch.float32)
        fn = (y_true * (1 - y_pred)).sum(dim=0).to(torch.float32)

        precision = tp / (tp + fp + self.epsilon)
        recall = tp / (tp + fn + self.epsilon)

        f1 = 2 * (precision * recall) / (precision + recall + self.epsilon)
        f1 = f1.clamp(min=self.epsilon, max=1 - self.epsilon)

        # unweighted mean F1 loss
        if self.weight is None:
            return 1 - f1.mean()

        # weighted mean F1 loss
        wm = torch.sum(f1 * self.weight) / torch.sum(self.weight)
        return 1 - wm



# modelrunner.py

from nkululeko.util import Util
from nkululeko import glob_conf


class Modelrunner:
    """
    Class to model one run
    """

    def __init__(self, df_train, df_test, feats_train, feats_test, run):
        """Constructor setting up the dataframes
        Args:
            df_train: train dataframe
            df_test: test dataframe
            feats_train: train features
            feats_train: test features
        """
        self.df_train, self.df_test, self.feats_train, self.feats_test = (
            df_train,
            df_test,
            feats_train,
            feats_test,
        )
        self.util = Util("modelrunner")
        self.run = run
        self.target = glob_conf.config["DATA"]["target"]
        # intialize a new model
        model_type = glob_conf.config["MODEL"]["type"]
        self._select_model(model_type)

    def do_epochs(self):
        # initialze results
        reports = []
        plot_epochs = self.util.config_val("PLOT", "epochs", False)
        only_test = self.util.config_val("MODEL", "only_test", False)
        # for all epochs
        for epoch in range(int(self.util.config_val("EXP", "epochs", 1))):
            if only_test:
                self.model.load(self.run, epoch)
                self.util.debug(f"reusing model: {self.model.store_path}")
                self.model.reset_test(self.df_test, self.feats_test)
            else:
                self.model.set_id(self.run, epoch)
                self.model.train()
            report = self.model.predict()
            report.set_id(self.run, epoch)
            plot_name = self.util.get_plot_name() + f"_{self.run}_{epoch:03d}_cnf"
            reports.append(report)
            self.util.debug(
                f"run: {self.run} epoch: {epoch}: result: "
                f"{reports[-1].get_result().get_test_result()}"
            )
            if plot_epochs:
                self.util.debug(f"plotting conf matrix to {plot_name}")
                report.plot_confmatrix(plot_name, epoch)
            store_models = self.util.config_val("MODEL", "save", False)
            plot_best_model = self.util.config_val("PLOT", "best_model", False)
            if (store_models or plot_best_model) and (
                not only_test
            ):  # in any case the model needs to be stored to disk.
                self.model.store()
        if not plot_epochs:
            # Do at least one confusion matrix plot
            self.util.debug(f"plotting confusion matrix to {plot_name}")
            reports[-1].plot_confmatrix(plot_name, epoch)
        return reports

    def _select_model(self, model_type):
        if model_type == "svm":
            from nkululeko.models.model_svm import SVM_model

            self.model = SVM_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "svr":
            from nkululeko.models.model_svr import SVR_model

            self.model = SVR_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "xgb":
            from nkululeko.models.model_xgb import XGB_model

            self.model = XGB_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "xgr":
            from nkululeko.models.model_xgr import XGR_model

            self.model = XGR_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "bayes":
            from nkululeko.models.model_bayes import Bayes_model

            self.model = Bayes_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "gmm":
            from nkululeko.models.model_gmm import GMM_model

            self.model = GMM_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "knn":
            from nkululeko.models.model_knn import KNN_model

            self.model = KNN_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "knn_reg":
            from nkululeko.models.model_knn_reg import KNN_reg_model

            self.model = KNN_reg_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "tree":
            from nkululeko.models.model_tree import Tree_model

            self.model = Tree_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "tree_reg":
            from nkululeko.models.model_tree_reg import Tree_reg_model

            self.model = Tree_reg_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "cnn":
            from nkululeko.models.model_cnn import CNN_model
            from nkululeko.models.model_cnn import CNN_model

            self.model = CNN_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "mlp":
            from nkululeko.models.model_mlp import MLP_model

            self.model = MLP_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "mlp_reg":
            from nkululeko.models.model_mlp_regression import MLP_Reg_model

            self.model = MLP_Reg_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        else:
            self.util.error(f"unknown model type: '{model_type}'")
        if self.util.exp_is_classification() and not self.model.is_classifier:
            self.util.error(
                "Experiment type set to classification but model type is not a classifier"
            )
        return self.model



""""
A predictor for emotional arousal.
Currently based on audEERING's emotional dimension model.
"""
from nkululeko.util import Util
from nkululeko.feature_extractor import FeatureExtractor
import ast
import nkululeko.glob_conf as glob_conf


class ArousalPredictor:
    """
    ArousalPredictor
    predicting arousal with the audEERING emotional dimension model

    """

    def __init__(self, df):
        self.df = df
        self.util = Util("arousalPredictor")

    def predict(self, split_selection):
        self.util.debug(f"predicting arousal for {split_selection} samples")
        feats_name = "_".join(ast.literal_eval(glob_conf.config["DATA"]["databases"]))
        self.feature_extractor = FeatureExtractor(
            self.df, ["auddim"], feats_name, split_selection
        )
        pred_df = self.feature_extractor.extract()
        pred_vals = pred_df.arousal * 1000
        return_df = self.df.copy()
        return_df["arousal_pred"] = pred_vals.astype("int") / 1000

        return return_df



import matplotlib.pyplot as plt
from sklearn.utils import resample
from nkululeko.util import Util 
import ast
import json
import numpy as np
import nkululeko.glob_conf as glob_conf
from sklearn.metrics import ConfusionMatrixDisplay
from sklearn.metrics import confusion_matrix
from sklearn.metrics import recall_score
from sklearn.metrics import accuracy_score
from sklearn.metrics import mean_squared_error
from sklearn.metrics import r2_score
from sklearn.metrics import classification_report
from scipy.stats import pearsonr
from nkululeko.result import Result
from nkululeko.reporting.report_item import ReportItem
import glob
import math

class Reporter:

    def __set_measure(self):
        if self.util.exp_is_classification():
            self.MEASURE = 'UAR'
            self.result.measure = self.MEASURE
            self.is_classification = True
        else:
            self.is_classification = False
            self.measure = self.util.config_val('MODEL', 'measure', 'mse')
            if self.measure == 'mse':
                self.MEASURE = 'MSE'
                self.result.measure = self.MEASURE
            elif self.measure == 'ccc':
                self.MEASURE = 'CCC'
                self.result.measure = self.MEASURE


    def __init__(self, truths, preds, run, epoch):
        """Initialization with ground truth und predictions vector"""
        self.util = Util('reporter')
        self.format = self.util.config_val('PLOT', 'format', 'png')
        self.truths = truths
        self.preds = preds
        self.result = Result(0, 0, 0, 'unknown')
        self.run = run
        self.epoch = epoch
        self.__set_measure()
        self.cont_to_cat = False
        if len(self.truths)>0 and len(self.preds)>0:
            if self.util.exp_is_classification():
                self.result.test = recall_score(self.truths, self.preds, average='macro')
                self.result.loss = 1 - accuracy_score(self.truths, self.preds)
            else:
                # regression experiment
                if self.measure == 'mse':
                    self.result.test = mean_squared_error(self.truths, self.preds)
                elif self.measure == 'ccc':
                    self.result.test = self.ccc(self.truths, self.preds)
                    if math.isnan(self.result.test):
                        self.util.debug(f'Truth: {self.truths}')
                        self.util.debug(f'Predict.: {self.preds}')
                        self.util.debug(f'Result is NAN: setting to -1')
                        self.result.test = -1
                else:
                    self.util.error(f'unknown measure: {self.measure}')

                # train and loss are being set by the model

    def set_id(self, run, epoch):
        """Make the report identifiable with run and epoch index"""
        self.run = run
        self.epoch = epoch

    def continuous_to_categorical(self):
        if self.cont_to_cat:
            return 
        self.cont_to_cat = True
        bins = ast.literal_eval(glob_conf.config['DATA']['bins'])
        self.truths = np.digitize(self.truths, bins)-1
        self.preds = np.digitize(self.preds, bins)-1

    def plot_confmatrix(self, plot_name, epoch): 
        if not self.util.exp_is_classification():
            self.continuous_to_categorical()
        self._plot_confmat(self.truths, self.preds, plot_name, epoch)

    def plot_per_speaker(self, result_df, plot_name, function):
        """ Plot a confusion matrix with the mode category per speakers
            Args:
                * result_df: a pandas dataframe with columns: preds, truths and speaker
        """
        speakers = result_df.speaker.unique()
        pred = np.zeros(0)
        truth = np.zeros(0)
        for s in speakers:
            s_df = result_df[result_df.speaker==s]
            mode = s_df.pred.mode().iloc[-1]
            mean = s_df.pred.mean()
            if function == 'mode':
                s_df.pred=mode
            elif function == 'mean':
                s_df.pred=mean
            else:
                self.util.error(f'unkown function {function}')
            pred = np.append(pred, s_df.pred.values)     
            truth = np.append(truth, s_df['truth'].values)
        if not (self.is_classification or self.cont_to_cat):
            bins = ast.literal_eval(glob_conf.config['DATA']['bins'])
            truth = np.digitize(truth, bins)-1
            pred = np.digitize(pred, bins)-1
        self._plot_confmat(truth, pred.astype('int'), plot_name, 0)            


    def _plot_confmat(self, truths, preds, plot_name, epoch): 
        # print(truths)
        # print(preds)
        fig_dir = self.util.get_path('fig_dir')
        labels = glob_conf.labels
        fig = plt.figure()  # figsize=[5, 5]
        uar = recall_score(truths, preds, average='macro')
        acc = accuracy_score(truths, preds)
        cm = confusion_matrix(truths, preds, normalize = None) #normalize must be one of {'true', 'pred', 'all', None}
        if cm.shape[0] != len(labels):
            self.util.error(f'mismatch between confmatrix dim ({cm.shape[0]}) and labels length ({len(labels)}: {labels})')
        try:
            disp = ConfusionMatrixDisplay(confusion_matrix=cm, display_labels=labels).plot(cmap='Blues')
        except ValueError:
            disp = ConfusionMatrixDisplay(confusion_matrix=cm, display_labels=list(labels).remove('neutral')).plot(cmap='Blues')

        reg_res = ''
        if not self.is_classification:
            reg_res = f', {self.MEASURE}: {self.result.test:.3f}'

        if epoch != 0:
            plt.title(f'Confusion Matrix, UAR: {uar:.3f}{reg_res}, Epoch: {epoch}')
        else:
            plt.title(f'Confusion Matrix, UAR: {uar:.3f}{reg_res}')

        plt.savefig(f'{fig_dir}{plot_name}.{self.format}')
        fig.clear()
        plt.close(fig)
        res_dir = self.util.get_path('res_dir')
        rpt = f'epoch: {epoch}, UAR: {uar}, ACC {acc}'
        file_name = f'{res_dir}{self.util.get_exp_name()}_conf.txt'
        with open(file_name, "w") as text_file:
            text_file.write(rpt)

    def print_results(self, epoch):
        """Print all evaluation values to text file"""
        res_dir = self.util.get_path('res_dir')
        file_name = f'{res_dir}{self.util.get_exp_name()}_{epoch}.txt'
        if self.util.exp_is_classification():
            data_type = self.util.config_val('DATA', 'type', 'False')
            labels = glob_conf.labels
            # if data_type == 'continuous' or data_type == 'continous':
            #     labels = glob_conf.labels
            # else:
            #     labels = glob_conf.label_encoder.classes_
            try:
                rpt = classification_report(self.truths, self.preds, target_names=labels, output_dict=True)
            except ValueError as e:
                self.util.debug('Reporter: caught a ValueError when trying to get classification_report: '+e)
                rpt = self.result.to_string()
            with open(file_name, "w") as text_file:
                c_ress = list(range(len(labels)))
                for i, l in enumerate(labels):
                    c_res = rpt[l]['f1-score']
                    c_ress[i] = float(f'{c_res:.3f}')
                self.util.debug(f'labels: {labels}')
                f1_per_class = f'result per class (F1 score): {c_ress}'
                self.util.debug(f1_per_class)
                rpt_str = f'{json.dumps(rpt)}\n{f1_per_class}'
                text_file.write(rpt_str)
                glob_conf.report.add_item(ReportItem('results', 'Result', rpt_str))

        else: # regression
            result = self.result.test
            r2 = r2_score(self.truths, self.preds)
            pcc = pearsonr(self.truths, self.preds)[0]
            measure = self.util.config_val('MODEL', 'measure', 'mse')
            with open(file_name, "w") as text_file:
                text_file.write(f'{measure}: {result:.3f}, r_2: {r2:.3f}, pcc {pcc:.3f}')

    def make_conf_animation(self, out_name):
        import imageio
        fig_dir = self.util.get_path('fig_dir')
        filenames =  glob.glob(fig_dir+f'{self.util.get_plot_name()}*_?_???_cnf.png')
        images = []
        for filename in filenames:
            images.append(imageio.imread(filename))
        fps = self.util.config_val('PLOT', 'fps', '1')
        try:
            imageio.mimsave(fig_dir+out_name, images, fps=int(fps))
        except RuntimeError as e:
            self.util.error('error writing anim gif: '+e)

    def get_result(self):
        return self.result


    def plot_epoch_progression(self, reports, out_name):
        fig_dir = self.util.get_path('fig_dir')
        results, losses, train_results = [], [], []
        for r in reports:
            results.append(r.get_result().test)
            losses.append(r.get_result().loss)
            train_results.append(r.get_result().train)

        # do a plot per run
        # scale the losses so they fit on the picture
        losses, results, train_results = np.asarray(losses), np.asarray(results), np.asarray(train_results)

        if np.all((results < 1)):
            # scale up values 
            results = results*100
            train_results = train_results*100
        if np.all((losses < 1)):
            # scale up values 
            losses = losses*100
        plt.figure(dpi=200)
        plt.plot(train_results, 'green', label='train set') 
        plt.plot(results, 'red', label='dev set')
        plt.plot(losses, 'grey', label='losses')
        plt.xlabel('epochs')
        plt.ylabel(f'{self.MEASURE}')
        plt.legend()
        plt.savefig(f'{fig_dir}{out_name}.{self.format}')
        plt.close()        

    @staticmethod
    def ccc(ground_truth, prediction):
        mean_gt = np.mean(ground_truth, 0)
        mean_pred = np.mean(prediction, 0)
        var_gt = np.var (ground_truth, 0)
        var_pred = np.var (prediction, 0)
        v_pred = prediction - mean_pred
        v_gt = ground_truth - mean_gt
        cor = sum (v_pred * v_gt) / (np.sqrt(sum(v_pred ** 2)) * np.sqrt(sum(v_gt ** 2)))
        sd_gt = np.std(ground_truth)
        sd_pred = np.std(prediction)
        numerator=2*cor*sd_gt*sd_pred
        denominator=var_gt+var_pred+(mean_gt-mean_pred)**2
        ccc = numerator/denominator
        return ccc


# feats_wav2vec2.py
# feat_types example = wav2vec2-large-robust-ft-swbd-300h
import os
from tqdm import tqdm
import pandas as pd
import torch
import torchaudio
from transformers import Wav2Vec2FeatureExtractor, Wav2Vec2Model
from nkululeko.feat_extract.featureset import Featureset
import nkululeko.glob_conf as glob_conf

class Wav2vec2(Featureset):
    """Class to extract wav2vec2 embeddings"""

    def __init__(self, name, data_df, feat_type):
        """Constructor. is_train is needed to distinguish from test/dev sets, because they use the codebook from the training"""
        super().__init__(name, data_df)
        cuda = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = self.util.config_val("MODEL", "device", cuda)
        self.model_initialized = False
        self.feat_type = feat_type

    def init_model(self):
        # load model
        self.util.debug("loading wav2vec model...")
        model_path = self.util.config_val(
            "FEATS", "wav2vec.model", f"facebook/{self.feat_type}"
        )
        self.processor = Wav2Vec2FeatureExtractor.from_pretrained(model_path)
        self.model = Wav2Vec2Model.from_pretrained(model_path).to(self.device)
        print(f"intialized Wav2vec model on {self.device}")
        self.model.eval()
        self.model_initialized = True

    def extract(self):
        """Extract the features or load them from disk if present."""
        store = self.util.get_path("store")
        storage = f"{store}{self.name}.pkl"
        extract = self.util.config_val("FEATS", "needs_feature_extraction", False)
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            if not self.model_initialized:
                self.init_model()
            self.util.debug(
                "extracting wav2vec2 embeddings, this might take a while..."
            )
            emb_series = pd.Series(index=self.data_df.index, dtype=object)
            length = len(self.data_df.index)
            for idx, (file, start, end) in enumerate(tqdm(self.data_df.index.to_list())):
                signal, sampling_rate = torchaudio.load(file,
                    frame_offset=int(start.total_seconds()*16000),
                    num_frames=int((end - start).total_seconds()*16000))
                assert sampling_rate == 16000, f"got {sampling_rate} instead of 16000"
                emb = self.get_embeddings(signal, sampling_rate, file)
                emb_series[idx] = emb
            print(f"emb_series shape: {emb_series.shape}")
            self.df = pd.DataFrame(emb_series.values.tolist(), 
                                   index=self.data_df.index)
            print(f"df shape: {self.df.shape}")
            self.df.to_pickle(storage)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "false"
            except KeyError:
                pass
        else:
            self.util.debug("reusing extracted wav2vec2 embeddings")
            self.df = pd.read_pickle(storage)
            if self.df.isnull().values.any():
                nanrows = self.df.columns[self.df.isna().any()].tolist()
                print(nanrows)
                self.util.error(
                    f"got nan: {self.df.shape} {self.df.isnull().sum().sum()}"
                )

    def get_embeddings(self, signal, sampling_rate, file):
        r"""Extract embeddings from raw audio signal."""
        try:
            with torch.no_grad():
                # run through processor to normalize signal
                # always returns a batch, so we just get the first entry
                # then we put it on the device
                y = self.processor(signal, sampling_rate=sampling_rate)
                y = y["input_values"][0]
                y = torch.from_numpy(y.reshape(1, -1)).to(self.device)
                # print(y.shape)
                # run through model
                # first entry contains hidden state
                y = self.model(y)[0]

                # pool result and convert to numpy
                y = torch.mean(y, dim=1)
                y = y.detach().cpu().numpy()
        except RuntimeError as re:
            print(str(re))
            self.util.error(f"couldn't extract file: {file}")
        print(f"y flattened shape: {y.ravel().shape}")
        return y.ravel()

    def extract_sample(self, signal, sr):
        self.init_model()
        feats = self.get_embeddings(signal, sr, "no file")
        return feats



""" test_predictor.py
    Predict targets from a model and save as csv file.

"""
import nkululeko.glob_conf as glob_conf
from nkululeko.util import Util
import pandas as pd
from nkululeko.data.dataset import Dataset
from nkululeko.feature_extractor import FeatureExtractor
from nkululeko.scaler import Scaler
import numpy as np
from sklearn.preprocessing import LabelEncoder


class Test_predictor():


    def __init__(self, model, orig_df, labenc, name):
        """Constructor setting up name and configuration"""
        self.model = model
        self.orig_df = orig_df
        self.label_encoder = labenc
        self.target = glob_conf.config['DATA']['target']
        self.util = Util('test_predictor')
        self.name = name

    def predict_and_store(self):
        label_data = self.util.config_val('DATA', 'label_data', False)
        if label_data:
            data = Dataset(label_data)
            data.load()
            data.prepare_labels()
            data_df = self.util.make_segmented_index(data.df)
            data_df.is_labeled = data.is_labeled
            
            featextractor = FeatureExtractor(data_df, label_data, '')
            feats_df = featextractor.extract()
            scale = self.util.config_val('FEATS', 'scale', False)
            labelenc = LabelEncoder()
            data_df[self.target] = labelenc.fit_transform(data_df[self.target])
#            data_df[self.target] = self.label_encoder.fit_transform(data_df[self.target])
            if scale: 
                self.scaler = Scaler(data_df, None, feats_df, None, scale)
                feats_df, _ = self.scaler.scale()
            self.model.set_testdata(data_df, feats_df)
            predictions = self.model.get_predictions()
            df = pd.DataFrame(index = data_df.index)
            df['speaker'] = data_df['speaker']
            df['gender'] = data_df['gender']
            df[self.target] = labelenc.inverse_transform(predictions.tolist())
            df.to_csv(self.name)
        else:
            predictions = self.model.get_predictions()
            #print(predictions)
            df = pd.DataFrame(index = self.orig_df.index)
            df['speaker'] = self.orig_df['speaker']
            df['gender'] = self.orig_df['gender']
            df[self.target] = self.label_encoder.inverse_transform(predictions)
            #df[self.target] = predictions
            df.to_csv(self.name)



# feats_trill.py
import os
import tensorflow as tf
from numpy.core.numeric import tensordot
from tqdm import tqdm
import pandas as pd
import audiofile as af
from nkululeko.util import Util
import nkululeko.glob_conf as glob_conf
from nkululeko.feat_extract.featureset import Featureset

# Import TF 2.X and make sure we're running eager.
assert tf.executing_eagerly()
import tensorflow_hub as hub


class TRILLset(Featureset):
    """A feature extractor for the Google TRILL embeddings"""

    """https://ai.googleblog.com/2020/06/improving-speech-representations-and.html"""

    # Initialization of the class
    def __init__(self, name, data_df):
        """
        Initialize the class with name, data and Util instance
        Also loads the model from hub

        :param name: Name of the class
        :type name: str
        :param data_df: Data of the class
        :type data_df: DataFrame
        :return: None
        """
        super().__init__(name, data_df)
        # Load the model from the configured path
        model_path = self.util.config_val(
            "FEATS",
            "trill.model",
            "https://tfhub.dev/google/nonsemantic-speech-benchmark/trill/3",
        )
        self.module = hub.load(model_path)

    def extract(self):
        store = self.util.get_path("store")
        storage = f"{store}{self.name}.pkl"
        extract = self.util.config_val("FEATS", "needs_feature_extraction", False)
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            self.util.debug("extracting TRILL embeddings, this might take a while...")
            emb_series = pd.Series(index=self.data_df.index, dtype=object)
            length = len(self.data_df.index)
            for idx, file in enumerate(tqdm(self.data_df.index.get_level_values(0))):
                emb = self.getEmbeddings(file)
                emb_series[idx] = emb
            self.df = pd.DataFrame(emb_series.values.tolist(), index=self.data_df.index)
            self.df.to_pickle(storage)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "false"
            except KeyError:
                pass
        else:
            self.util.debug("reusing extracted TRILL embeddings")
            self.df = pd.read_pickle(storage)

    def embed_wav(self, wav):
        if len(wav.shape) > 1:
            wav = tf.reduce_mean(wav, axis=0)

        emb_dict = self.module(samples=wav, sample_rate=tf.constant(16000))
        return emb_dict["embedding"]

    def getEmbeddings(self, file):
        wav = af.read(file)[0]
        wav = tf.convert_to_tensor(wav)
        emb_short = self.embed_wav(wav)
        # you get one embedding per frame, we use the mean for all the frames
        emb_short = emb_short.numpy().mean(axis=0)
        return emb_short



"""
A predictor for emotional dominance.
Currently based on audEERING's emotional dimension model.
"""

from nkululeko.util import Util
from nkululeko.feature_extractor import FeatureExtractor
import ast
import nkululeko.glob_conf as glob_conf


class DominancePredictor:
    """
    DominancePredictor
    predicting dominance with the audEERING emotional dimension model
    """

    def __init__(self, df):
        self.df = df
        self.util = Util("dominancePredictor")

    def predict(self, split_selection):
        self.util.debug(f"predicting dominance for {split_selection} samples")
        feats_name = "_".join(ast.literal_eval(glob_conf.config["DATA"]["databases"]))
        self.feature_extractor = FeatureExtractor(
            self.df, ["auddim"], feats_name, split_selection
        )
        pred_df = self.feature_extractor.extract()
        pred_vals = pred_df.dominance * 1000
        return_df = self.df.copy()
        return_df["dominance_pred"] = pred_vals.astype("int") / 1000

        return return_df



from itertools import combinations
import math
import numpy as np
import pandas as pd

def check_na(a):
    if np.isnan(a).any():
        count = np.count_nonzero(np.isnan(a))
        print(f'WARNING: got {count} Nans (of {len(a)}), setting to 0')
        a[np.isnan(a)] = 0
        return a 
    else:
        return a

def cohen_d(d1, d2):
    """
    Compute Cohen's d from two distributions of real valued arrays
    Args:
        d1: one array
        d2: the other array
    Returns:
        Cohen's d with precision 3
    """
    # Checks:
    d1 = check_na(d1)
    d2 = check_na(d2)
    # calculate the size of samples
    n1, n2 = len(d1), len(d2)
    # calculate the variance of the samples
    s1, s2 = np.var(d1, ddof=1), np.var(d2, ddof=1)
    # calculate the pooled standard deviation
    s = math.sqrt(((n1 - 1) * s1 + (n2 - 1) * s2) / (n1 + n2 - 2))
    # calculate the means of the samples
    u1, u2 = np.mean(d1), np.mean(d2)
    # calculate the effect size
    if math.isnan(s) or s == 0:
        return -1
    return (int(1000 * np.abs((u1 - u2)) / s)) / 1000


def all_combinations(items_list):
    result = []
    for i in range(1, len(items_list) + 1):
        for combo in combinations(items_list, i):
            if len(combo) == 2:
                result.append(list(combo))
    return result


def get_effect_size(df, target, variable):
    """
    Get the effect size as Cohen's D from a real numbered variable on a categorical target.
    Args:
        df: a pd.Dataframe with at least target and variable as columns
        target: the categorical target, e.g. emotion
        variable: the real numbered variable that might have an effect, e.g. SNR
    Returns:
        The categories with the maximal Cohen's d based on the variable as a string
        The Cohen's d for this combination
    """
    categories = df[target].unique()
    cats = {}
    for c in categories:
        cats[c] = df[df[target] == c][variable].values
    combos = all_combinations(categories)
    results = {}
    for combo in combos:
        one = combo[0]
        other = combo[1]
        results[f"{one}-{other}"] = cohen_d(cats[one], cats[other])
    max_cat = max(results, key=results.get)
    return max_cat, results[max_cat]



# scaler.py


from sklearn.preprocessing import StandardScaler, RobustScaler
import pandas as pd
from nkululeko.util import Util

class Scaler:
    """
        class to normalize speech parameters   
    """
    
    def __init__(self, train_data_df, test_data_df, train_feats, test_feats, scaler_type):
        '''
        Initializer.

                Parameters:
                        train_data_df (pd.DataFrame): The training dataframe with speakers.
                            only needed for speaker normalization
                        test_data_df (pd.DataFrame): The test dataframe with speakers
                            only needed for speaker normalization
                        train_feats (pd.DataFrame): The train features dataframe 
                        test_feats (pd.DataFrame): The test features dataframe (can be None) 
        '''
        self.util = Util('scaler')
        if scaler_type == 'standard':
            self.scaler = StandardScaler()
        elif scaler_type == 'robust':
            self.scaler = RobustScaler()
        elif scaler_type == 'speaker':
            self.scaler = StandardScaler()
        else:
            self.util.error('unknown scaler: '+scaler_type)
        self.scaler_type = scaler_type
        self.feats_train = train_feats
        self.data_train = train_data_df
        self.feats_test = test_feats
        self.data_test = test_data_df

    def scale(self):
        '''
        Actually scales/normalizes.

                Returns:
                        train_feats (pd.DataFrame): The scaled train features dataframe 
                        test_feats (pd.DataFrame): The scaled test features dataframe (can be None) 
        '''
        if self.scaler_type != 'speaker':
            self.util.debug('scaling features based on training set')
            return self.scale_all()
        else:
            self.util.debug('scaling features per speaker based on training')
            return self.speaker_scale()

    def scale_all(self):
        self.scaler.fit(self.feats_train.values)
        self.feats_train = self.scale_df(self.feats_train)
        if self.feats_test is not None:
            self.feats_test = self.scale_df(self.feats_test)
        return self.feats_train, self.feats_test

    def scale_df(self, df):
        scaled_features = self.scaler.fit_transform(df.values)
        df = pd.DataFrame(scaled_features, index=df.index, columns=df.columns)
        return df 

    def speaker_scale(self):
        self.feats_train = self.speaker_scale_df(self.data_train, self.feats_train)
        if self.feats_test is not None:
            self.feats_test = self.speaker_scale_df(self.data_test, self.feats_test)
        return [self.feats_train, self.feats_test]


    def speaker_scale_df(self, df, feats_df):
        for speaker in df['speaker'].unique():
            indices = df.loc[df['speaker'] == speaker].index
            feats_df.loc[indices, :] = self.scaler.fit_transform(feats_df.loc[indices, :]) 
        return feats_df


# global_config.py


def init_config(config_obj):
    global config
    config = config_obj


def set_label_encoder(encoder):
    global label_encoder
    label_encoder = encoder


def set_util(util_obj):
    global util
    util = util_obj

def set_report(report_obj):
    global report
    report = report_obj

def set_labels(labels_obj):
    global labels
    labels = labels_obj